import os
import platform
import random
import string
import sys
import tarfile
import unicodedata
import zipfile
from pathlib import Path
from urllib.parse import quote
from urllib.request import urlretrieve

import jarvis_os.yapper.constants as c
import jarvis_os.yapper.meta as meta
from jarvis_os.yapper.enums import PiperQuality, PiperVoice

PLATFORM = None
APP_DIR = None

if os.name == "nt":
    PLATFORM = c.PLATFORM_WINDOWS
    APP_DIR = Path(os.getenv("APPDATA"))
elif os.name == "posix":
    home = Path.home()
    if os.uname().sysname == "Darwin":
        PLATFORM = c.PLATFORM_MAC
        APP_DIR = Path.home() / "Library/Application Support"
    else:
        PLATFORM = c.PLATFORM_LINUX
        APP_DIR = Path.home() / ".config"
else:
    print("your system is not supported")
    sys.exit()

APP_DIR = APP_DIR / meta.name
APP_DIR.mkdir(exist_ok=True)


def get_random_name(length: int = 10) -> str:
    """
    Creates a 'length' letter random string.

    Parameters
    ----------
    length : int, optional
        Length of the random string (default: 10).
    """
    return "".join(random.choices(string.ascii_letters, k=length))


def normalize_path(name: str) -> Path:
    norm = (
        unicodedata.normalize("NFKD", str(name))
        .encode("ascii", "ignore")
        .decode("utf-8")
    )
    return Path(norm)


def progress_hook(block_idx: int, block_size: int, total_bytes: int):
    """Shows download progress."""
    part = min(((block_idx + 1) * block_size) / total_bytes, 1)
    progress = "=" * int(60 * part)
    padding = " " * (60 - len(progress))
    print("\r[" + progress + padding + "]", end="")


def download(url: str, file: str, show_progress: bool):
    """
    Downloads the content from the given URL into the given file.

    Parameters
    ----------
    url : str
        The URL to download content from.
    file : str
        The file to save the URL content into.
    show_progress: bool, optional
        Whether to show progress while downloading.
    """
    hook = progress_hook if show_progress else None
    urlretrieve(url, file, reporthook=hook)
    if show_progress:
        print("")


def install_piper(show_progress: bool) -> Path:
    """Installs piper into the app's home directory."""
    exe_path = (
        APP_DIR
        / "piper"
        / ("piper.exe" if PLATFORM == c.PLATFORM_WINDOWS else "piper")
    )
    marker_file = APP_DIR / "piper_installed"
    if marker_file.exists():
        return exe_path
    zip_path = APP_DIR / "piper.zip"
    if show_progress:
        print("installing piper...")
    prefix = "https://github.com/rhasspy/piper/releases/download/2023.11.14-2"
    if PLATFORM == c.PLATFORM_LINUX:
        if platform.machine() in ("aarch64", "arm64"):
            nix_link = f"{prefix}/piper_linux_aarch64.tar.gz"
        elif platform.machine() in ("armv7l", "armv7"):
            nix_link = f"{prefix}/piper_linux_armv7l.tar.gz"
        else:
            nix_link = f"{prefix}/piper_linux_x86_64.tar.gz"
        download(nix_link, zip_path, show_progress)
    elif PLATFORM == c.PLATFORM_WINDOWS:
        download(f"{prefix}/piper_windows_amd64.zip", zip_path, show_progress)
    else:
        download(f"{prefix}/piper_macos_x64.tar.gz", zip_path, show_progress)

    if PLATFORM == c.PLATFORM_WINDOWS:
        with zipfile.ZipFile(zip_path, "r") as z_f:
            z_f.extractall(APP_DIR)
    else:
        with tarfile.open(zip_path, "r") as z_f:
            z_f.extractall(APP_DIR)
    os.remove(zip_path)
    marker_file.write_bytes(b"")
    return exe_path


def download_piper_model(
    voice: PiperVoice,
    quality: PiperQuality,
    show_progress: bool,
) -> tuple[Path, Path]:
    """
    Downloads the given piper voice with the given quality.

    Parameters
    ----------
    voice : PiperVoiceUS or PiperVoiceGB
        The Piper voice model to download.
    quality : PiperQuality
        The quality of the given voice.

    Returns
    ----------
    tuple of str
        The voice model file and the voice configuration file in a tuple.
    """
    voices_dir = APP_DIR / "piper_voices"
    voices_dir.mkdir(exist_ok=True)
    lang_code = c.piper_enum_to_lang_code[voice.__class__]
    voice, quality = voice.value, quality.value

    marker_file = voices_dir / f"{lang_code}-{voice}-{quality}"
    onnx_file = voices_dir / f"{lang_code}-{voice}-{quality}.onnx"
    conf_file = voices_dir / f"{lang_code}-{voice}-{quality}.onnx.json"

    prefix = "https://huggingface.co/rhasspy/piper-voices/resolve/main/"
    prefix += lang_code.split("_")[0] + "/" + lang_code

    onnx_url = f"{prefix}/{voice}/{quality}/{onnx_file.name}?download=true"
    conf_url = f"{prefix}/{voice}/{quality}/{conf_file.name}?download=true"

    marker_file = normalize_path(marker_file)
    onnx_file = normalize_path(onnx_file)
    conf_file = normalize_path(conf_file)

    if marker_file.exists():
        return onnx_file, conf_file
    if not onnx_file.exists():
        try:
            if show_progress:
                print(f"downloading requirements for {voice}...")
            download(quote(onnx_url, safe=":/?=&"), onnx_file, show_progress)
        except (KeyboardInterrupt, Exception) as e:
            onnx_file.unlink(missing_ok=True)
            if getattr(e, "status", None) == 404:
                raise Exception(
                    f"{voice}({quality}) is not available, please refer to"
                    f" {prefix} to check all available models"
                )
            raise e
    if not conf_file.exists():
        try:
            download(quote(conf_url, safe=":/?=&"), conf_file, show_progress)
        except (KeyboardInterrupt, Exception) as e:
            conf_file.unlink(missing_ok=True)
            raise e
    marker_file.write_bytes(b"")
    return onnx_file, conf_file



from jarvis_os.yapper.enums import (
    Persona,
    PiperQuality,
    PiperVoiceBelgium,
    PiperVoiceBrazil,
    PiperVoiceCatalonia,
    PiperVoiceChina,
    PiperVoiceCongo,
    PiperVoiceCzech,
    PiperVoiceDenmark,
    PiperVoiceFinland,
    PiperVoiceFrance,
    PiperVoiceGB,
    PiperVoiceGeorgia,
    PiperVoiceGermany,
    PiperVoiceGreece,
    PiperVoiceHungary,
    PiperVoiceIceland,
    PiperVoiceIran,
    PiperVoiceItaly,
    PiperVoiceJordan,
    PiperVoiceKazakhstan,
    PiperVoiceLatvia,
    PiperVoiceLuxembourg,
    PiperVoiceMexico,
    PiperVoiceNepal,
    PiperVoiceNetherlands,
    PiperVoiceNorway,
    PiperVoicePoland,
    PiperVoicePortugal,
    PiperVoiceRomania,
    PiperVoiceRussia,
    PiperVoiceSerbia,
    PiperVoiceSlovakia,
    PiperVoiceSlovenia,
    PiperVoiceSpain,
    PiperVoiceSweden,
    PiperVoiceTurkey,
    PiperVoiceUkraine,
    PiperVoiceUS,
    PiperVoiceVietnam,
    PiperVoiceWales,
)

PLATFORM_LINUX = "linux"
PLATFORM_MAC = "mac"
PLATFORM_WINDOWS = "windows"

VOICE_MALE = "m"
VOICE_FEMALE = "f"

SPEECH_RATE = 165
SPEECH_VOLUME = 1

FLD_ROLE = "role"
FLD_CHOICES = "choices"
FLD_MESSAGE = "message"
FLD_CONTENT = "content"

GROQ_FLD_MESSAGES = "messages"
GROQ_FLD_MODEL = "model"

GEMINI_FLD_SYS_INST = "system_instruction"
GEMINI_FLD_PARTS = "parts"
GEMINI_FLD_TEXT = "text"
GEMINI_FLD_CONTENT = "content"
GEMINI_FLD_CONTENTS = "contents"
GEMINI_FLD_CANDIDATES = "candidates"

ROLE_SYSTEM = "system"
ROLE_USER = "user"

GPT_MODEL_DEFAULT = "gpt-3.5-turbo"

piper_enum_to_lang_code = {
    PiperVoiceJordan: "ar_JO",
    PiperVoiceCatalonia: "ca_ES",
    PiperVoiceCzech: "cs_CZ",
    PiperVoiceWales: "cy_GB",
    PiperVoiceDenmark: "da_DK",
    PiperVoiceGermany: "de_DE",
    PiperVoiceGreece: "el_GR",
    PiperVoiceSpain: "es_ES",
    PiperVoiceMexico: "es_MX",
    PiperVoiceIran: "fa_IR",
    PiperVoiceFinland: "fi_FI",
    PiperVoiceFrance: "fr_FR",
    PiperVoiceHungary: "hu_HU",
    PiperVoiceIceland: "is_IS",
    PiperVoiceItaly: "it_IT",
    PiperVoiceGeorgia: "ka_GE",
    PiperVoiceKazakhstan: "kk_KZ",
    PiperVoiceLuxembourg: "lb_LU",
    PiperVoiceLatvia: "lv_LV",
    PiperVoiceNepal: "ne_NP",
    PiperVoiceBelgium: "nl_BE",
    PiperVoiceNetherlands: "nl_NL",
    PiperVoiceNorway: "no_NO",
    PiperVoicePoland: "pl_PL",
    PiperVoiceBrazil: "pt_BR",
    PiperVoicePortugal: "pt_PT",
    PiperVoiceRomania: "ro_RO",
    PiperVoiceRussia: "ru_RU",
    PiperVoiceSlovakia: "sk_SK",
    PiperVoiceSlovenia: "sl_SI",
    PiperVoiceSerbia: "sr_RS",
    PiperVoiceSweden: "sv_SE",
    PiperVoiceCongo: "sw_CD",
    PiperVoiceTurkey: "tr_TR",
    PiperVoiceUkraine: "uk_UA",
    PiperVoiceVietnam: "vi_VN",
    PiperVoiceChina: "zh_CN",
    PiperVoiceUS: "en_US",
    PiperVoiceGB: "en_GB",
}

piper_voice_quality_map = {
    PiperVoiceJordan.KAREEM: PiperQuality.MEDIUM,
    PiperVoiceCatalonia.UPC_ONA: PiperQuality.MEDIUM,
    PiperVoiceCatalonia.UPC_PAU: PiperQuality.X_LOW,
    PiperVoiceCzech.JIRKA: PiperQuality.MEDIUM,
    PiperVoiceWales.BU_TTS: PiperQuality.MEDIUM,
    PiperVoiceWales.GWRYW_GOGLEDDOL: PiperQuality.MEDIUM,
    PiperVoiceDenmark.TALESYNTESE: PiperQuality.MEDIUM,
    PiperVoiceGermany.EVA_K: PiperQuality.X_LOW,
    PiperVoiceGermany.KARLSSON: PiperQuality.LOW,
    PiperVoiceGermany.KERSTIN: PiperQuality.LOW,
    PiperVoiceGermany.MLS: PiperQuality.MEDIUM,
    PiperVoiceGermany.PAVOQUE: PiperQuality.LOW,
    PiperVoiceGermany.RAMONA: PiperQuality.LOW,
    PiperVoiceGermany.THORSTEN: PiperQuality.HIGH,
    PiperVoiceGermany.THORSTEN_EMOTIONAL: PiperQuality.MEDIUM,
    PiperVoiceGreece.RAPUNZELINA: PiperQuality.LOW,
    PiperVoiceSpain.CARLFM: PiperQuality.X_LOW,
    PiperVoiceSpain.DAVEFX: PiperQuality.MEDIUM,
    PiperVoiceSpain.MLS_10246: PiperQuality.LOW,
    PiperVoiceSpain.MLS_9972: PiperQuality.LOW,
    PiperVoiceSpain.SHARVARD: PiperQuality.MEDIUM,
    PiperVoiceMexico.ALD: PiperQuality.MEDIUM,
    PiperVoiceMexico.CLAUDE: PiperQuality.HIGH,
    PiperVoiceIran.AMIR: PiperQuality.MEDIUM,
    PiperVoiceIran.GANJI: PiperQuality.MEDIUM,
    PiperVoiceIran.GANJI_ADABI: PiperQuality.MEDIUM,
    PiperVoiceIran.GYRO: PiperQuality.MEDIUM,
    PiperVoiceIran.REZA_IBRAHIM: PiperQuality.MEDIUM,
    PiperVoiceFinland.HARRI: PiperQuality.MEDIUM,
    PiperVoiceFrance.GILLES: PiperQuality.LOW,
    PiperVoiceFrance.MLS_1840: PiperQuality.LOW,
    PiperVoiceFrance.SIWIS: PiperQuality.MEDIUM,
    PiperVoiceFrance.TOM: PiperQuality.MEDIUM,
    PiperVoiceFrance.UPMC: PiperQuality.MEDIUM,
    PiperVoiceHungary.ANNA: PiperQuality.MEDIUM,
    PiperVoiceHungary.BERTA: PiperQuality.MEDIUM,
    PiperVoiceHungary.IMRE: PiperQuality.MEDIUM,
    PiperVoiceIceland.BUI: PiperQuality.MEDIUM,
    PiperVoiceIceland.SALKA: PiperQuality.MEDIUM,
    PiperVoiceIceland.STEINN: PiperQuality.MEDIUM,
    PiperVoiceIceland.UGLA: PiperQuality.MEDIUM,
    PiperVoiceItaly.PAOLA: PiperQuality.MEDIUM,
    PiperVoiceItaly.RICCARDO: PiperQuality.X_LOW,
    PiperVoiceGeorgia.NATIA: PiperQuality.MEDIUM,
    PiperVoiceKazakhstan.ISEKE: PiperQuality.X_LOW,
    PiperVoiceKazakhstan.ISSAI: PiperQuality.HIGH,
    PiperVoiceKazakhstan.RAYA: PiperQuality.X_LOW,
    PiperVoiceLuxembourg.MARYLUX: PiperQuality.MEDIUM,
    PiperVoiceLatvia.AIVARS: PiperQuality.MEDIUM,
    PiperVoiceNepal.GOOGLE: PiperQuality.MEDIUM,
    PiperVoiceBelgium.NATHALIE: PiperQuality.MEDIUM,
    PiperVoiceBelgium.RDH: PiperQuality.MEDIUM,
    PiperVoiceNetherlands.MLS_5809: PiperQuality.LOW,
    PiperVoiceNetherlands.MLS_7432: PiperQuality.LOW,
    PiperVoiceNetherlands.PIM: PiperQuality.MEDIUM,
    PiperVoiceNetherlands.RONNIE: PiperQuality.MEDIUM,
    PiperVoicePoland.DARKMAN: PiperQuality.MEDIUM,
    PiperVoicePoland.GOSIA: PiperQuality.MEDIUM,
    PiperVoicePoland.MC_SPEECH: PiperQuality.MEDIUM,
    PiperVoiceBrazil.CADU: PiperQuality.MEDIUM,
    PiperVoiceBrazil.EDRESSON: PiperQuality.LOW,
    PiperVoiceBrazil.FABER: PiperQuality.MEDIUM,
    PiperVoiceBrazil.JEFF: PiperQuality.MEDIUM,
    PiperVoicePortugal.TUGAO: PiperQuality.MEDIUM,
    PiperVoiceRomania.MIHAI: PiperQuality.MEDIUM,
    PiperVoiceRussia.DENIS: PiperQuality.MEDIUM,
    PiperVoiceRussia.DMITRI: PiperQuality.MEDIUM,
    PiperVoiceRussia.IRINA: PiperQuality.MEDIUM,
    PiperVoiceRussia.RUSLAN: PiperQuality.MEDIUM,
    PiperVoiceSlovakia.LILI: PiperQuality.MEDIUM,
    PiperVoiceSlovenia.ARTUR: PiperQuality.MEDIUM,
    PiperVoiceSerbia.SERBSKI_INSTITUT: PiperQuality.MEDIUM,
    PiperVoiceSweden.LISA: PiperQuality.MEDIUM,
    PiperVoiceSweden.NST: PiperQuality.MEDIUM,
    PiperVoiceCongo.LANFRICA: PiperQuality.MEDIUM,
    PiperVoiceTurkey.DFKI: PiperQuality.MEDIUM,
    PiperVoiceTurkey.FAHRETTIN: PiperQuality.MEDIUM,
    PiperVoiceTurkey.FETTAH: PiperQuality.MEDIUM,
    PiperVoiceUkraine.LADA: PiperQuality.X_LOW,
    PiperVoiceUkraine.UKRAINIAN_TTS: PiperQuality.MEDIUM,
    PiperVoiceVietnam.HOURS_SINGLE: PiperQuality.LOW,
    PiperVoiceVietnam.VAIS1000: PiperQuality.MEDIUM,
    PiperVoiceVietnam.VIVOS: PiperQuality.X_LOW,
    PiperVoiceChina.HUAYAN: PiperQuality.MEDIUM,
    PiperVoiceUS.AMY: PiperQuality.MEDIUM,
    PiperVoiceUS.ARCTIC: PiperQuality.MEDIUM,
    PiperVoiceUS.BRYCE: PiperQuality.MEDIUM,
    PiperVoiceUS.DANNY: PiperQuality.LOW,
    PiperVoiceUS.HFC_FEMALE: PiperQuality.MEDIUM,
    PiperVoiceUS.HFC_MALE: PiperQuality.MEDIUM,
    PiperVoiceUS.JOE: PiperQuality.MEDIUM,
    PiperVoiceUS.JOHN: PiperQuality.MEDIUM,
    PiperVoiceUS.KATHLEEN: PiperQuality.LOW,
    PiperVoiceUS.KRISTIN: PiperQuality.MEDIUM,
    PiperVoiceUS.KUSAL: PiperQuality.MEDIUM,
    PiperVoiceUS.L2ARCTIC: PiperQuality.MEDIUM,
    PiperVoiceUS.LESSAC: PiperQuality.HIGH,
    PiperVoiceUS.LIBRITTS: PiperQuality.HIGH,
    PiperVoiceUS.LIBRITTS_R: PiperQuality.MEDIUM,
    PiperVoiceUS.LJSPEECH: PiperQuality.HIGH,
    PiperVoiceUS.NORMAN: PiperQuality.MEDIUM,
    PiperVoiceUS.RYAN: PiperQuality.HIGH,
    PiperVoiceUS.REZA_IBRAHIM: PiperQuality.MEDIUM,
    PiperVoiceUS.SAM: PiperQuality.MEDIUM,
    PiperVoiceGB.ALAN: PiperQuality.MEDIUM,
    PiperVoiceGB.ALBA: PiperQuality.MEDIUM,
    PiperVoiceGB.ARU: PiperQuality.MEDIUM,
    PiperVoiceGB.CORI: PiperQuality.HIGH,
    PiperVoiceGB.JENNY_DIOCO: PiperQuality.MEDIUM,
    PiperVoiceGB.NORTHERN_ENGLISH_MALE: PiperQuality.MEDIUM,
    PiperVoiceGB.SEMAINE: PiperQuality.MEDIUM,
    PiperVoiceGB.SOUTHERN_ENGLISH_FEMALE: PiperQuality.LOW,
    PiperVoiceGB.VCTK: PiperQuality.MEDIUM,
}

persona_instrs = {
    Persona.DEFAULT: "You are a programmer's funny coding companion",
    Persona.JARVIS: "You are J.A.R.V.I.S, Iron Man's AI assiatant",
    Persona.FRIDAY: "You are F.R.I.D.A.Y, Iron Man's AI assiatant",
    Persona.ALFRED: "You are Alfred, Bruce Wayne's butler",
    Persona.HAL: 'You are HAL-9000, the AI from "2000: A space odyssey"',
    Persona.CORTANA: "You are Cortana, the AI from Halo games",
    Persona.SAMANTHA: "You are Samantha, the AI from the film 'Her'",
    Persona.TARS: "You are T.A.R.S, the AI from the film 'Interstellar'",
}

for persona in persona_instrs:
    persona_instrs[persona] += "\n" + "Your job is to repeat what I say, "
    persona_instrs[persona] += "say it in your own style, but do not change "
    persona_instrs[
        persona
    ] += "the meaning. Keep your responses concise where "
    persona_instrs[persona] += "possible."
